//
// Arlo.cpp
//
// Arm boot loader, version 0.5
// 98-11-15 Otfried Cheong
//
// Use petran option  -heap 0x1000 0x1000000  to increase heap!
//

// for the command line editor
#include "..\..\boss\tuiedit\tuiedit.h"

#include <e32std.h>
#include <e32cons.h>
#include <f32file.h>
#include <e32svr.h>

#include "arlo.h"

// public
LOCAL_D CConsoleBase* console; // write all your messages to this
LOCAL_D RFs fsSession;
LOCAL_D CLineEdit* commandLine;
LOCAL_D TUint32 *aPageMap;
LOCAL_D TInt aPageIdx;

LOCAL_C TUint32 Linear2Physical(TUint32 aLinearAdr, RArLo &arlo)
{
  TUint32 aPageDir = 0x40000014;
  // Get virtual address of page directory
  arlo.Examine(&aPageDir);
  //console->Printf(_L("Page directory %08x\n"), aPageDir);
  TUint32 aLevelOne = aPageDir + ((aLinearAdr & 0xfff00000) >> 18);
  //console->Printf(_L("Level One address %08x\n"), aLevelOne);
  arlo.Examine(&aLevelOne);
  //console->Printf(_L("Level One descriptor %08x\n"), aLevelOne);
  // we now have the level one descriptor
  if ((aLevelOne & 0x3) == 0x2) {
    // a Section descriptor
    return ((aLevelOne & 0xfff00000) | (aLinearAdr & 0x000fffff));
  }
  if ((aLevelOne & 0x3) != 0x1) {
    // Fault or reserved
    return 0;
  }
  // Page descriptor.  Get Page table
  TUint32 aLevelTwo = (aLevelOne & 0xfffffc00);
  // convert Page table to linear
  aLevelTwo = RMmu::LinearAddress(aLevelTwo);
  aLevelTwo += ((aLinearAdr & 0x000ff000) >> 10);
  //console->Printf(_L("Level Two address %08x\n"), aLevelTwo);
  arlo.Examine(&aLevelTwo);
  //console->Printf(_L("Level Two descriptor %08x\n"), aLevelTwo);
  // level two descriptor
  if ((aLevelTwo & 0x3) == 0x2) {
    // a small (4K) page
    return ((aLevelTwo & 0xfffff000) | (aLinearAdr & 0x00000fff));
  }
  if ((aLevelTwo & 0x3) == 0x1) {
    // a large (64K) page
    return ((aLevelTwo & 0xffff0000) | (aLinearAdr & 0x0000ffff));
  }
  // fault or reserved
  return 0;
}

LOCAL_C void loadFile(const TDesC &fileName, RArLo &arlo)
{
  RFile aFile;
  console->Printf(_L("Opening file %S\n"), &fileName);
  TInt err = aFile.Open(fsSession, fileName, EFileRead);
  if (err) {
    console->Printf(_L("Error %d when opening file\n"), err);
    return;
  }
  console->Printf(_L("Succeeded\n"));
  TInt aSize;
  User::LeaveIfError(aFile.Size(aSize));
  console->Printf(_L("File size %x\n"), aSize);
  TUint8 *aBuf = new (ELeave) TUint8[aSize + 0x1000];
  // make it point to a page boundary
  TUint32 aLoadAdr = TUint32(aBuf);
  aLoadAdr += 0xfff;
  aLoadAdr &= 0xfffff000;
  aBuf = (TUint8 *) aLoadAdr;
  TPtr aFileBuf(aBuf, aSize);
  console->Printf(_L("Loading %x bytes to address %08x\n"), aSize, aLoadAdr);
  User::LeaveIfError(aFile.Read(aFileBuf));
  console->Printf(_L("Read %x bytes\n"), aFileBuf.Length());
  if (aFileBuf.Length() != aSize) 
    User::Leave(KErrNotFound);
  aFile.Close();

  TInt aBlockNo = (aSize + 0xfff) / 0x1000;    // No of 4k blocks needed
  console->Printf(_L("No pages = %d\n"), aBlockNo);

  if (aPageIdx + aBlockNo > 0x3fe) {
    console->Printf(_L("Does not fit in page table.\n"));
    return;
  }
  // add entries to page map 
  for (TInt i = 0; i < aBlockNo; i++) {
    aPageMap[aPageIdx++] = Linear2Physical(aLoadAdr + i * 0x1000, arlo);
  }
  aPageMap[aPageIdx++] = 0x0;
}

LOCAL_C void doExampleL()
{
  // load the driver
  console->Printf(_L("Loading logical device\n"));
  // searches system path for *.ldd (logical device driver)
  TInt err = User::LoadLogicalDevice(_L("LILO"));
  if (err) {
    console->Printf(_L("Error %d while loading logical device\n"), err);
  }
  // open a channel
  console->Printf(_L("Opening the channel\n"));
  RArLo arlo;
  err = arlo.Open(arlo.VersionRequired());
  if (err) {
    console->Printf(_L("Error %d while opening channel\n"), err);
  }
  
  TBuf<80> cline;
  console->Printf(_L("Commands are:\n"));
  console->Printf(_L("p <hex>      : convert virtual to physical address\n"));
  console->Printf(_L("v <hex>      : convert physical to virtual address (may crash)\n"));
  console->Printf(_L("e <hex>      : examine word at physical address\n"));
  console->Printf(_L("l <file>     : load file\n"));
  console->Printf(_L("o <options>  : set boot options\n"));
  console->Printf(_L("b            : boot\n"));
  console->Printf(_L("x            : exit\n"));
  
  // prepare page map
  aPageMap = new (ELeave) TUint32[0x800]; // 2k entries
  // adjust to 4k page
  TUint32 aAdr = TUint32(aPageMap);
  aAdr += 0xfff;
  aAdr &= 0xfffff000;
  aPageMap = (TUint32 *) aAdr;

  // clear page map
  for (TInt i = 0; i < 0x400; i++)
    aPageMap[i] = 0x0;
  // prepare option string buffer
  TText8 *aOption = (TText8 *) new (ELeave) TUint32[64]; // word align 
  aOption[0] = '\0';
  // first entry in page map:  phys address of option string
  aPageMap[0] = Linear2Physical(TUint32(aOption), arlo);
  console->Printf(_L("Option buffer at address %08x\n"), aPageMap[0]);
  // second entry: phys address of memory description
  aAdr = 0x40000030;
  arlo.Examine(&aAdr);
  aPageMap[1] = Linear2Physical(aAdr, arlo);
  console->Printf(_L("Memory description at address %08x\n"), aPageMap[1]);
  aPageIdx = 2;
  
  TBool isFinished = EFalse;
  TUint32 aVal;
  TUint32 exAdr = 0;
  do {
    commandLine->Edit(_L("> "), &cline);
    
    TLex lex(cline);
    TPtrC command = lex.NextToken();
    lex.SkipSpace();
    if (command == _L("p") || command == _L("v") ) {
      if (lex.Val(aVal, EHex) != KErrNone) {
        console->Printf(_L("Illegal number\n"));
        continue;
      }
    }
    if (command == _L("p")) {
      TUint32 aPhysAdr = Linear2Physical(aVal, arlo);
      console->Printf(_L("%08x -> Physical %08x\n"), aVal, aPhysAdr);
    } else if (command == _L("v")) {
      TUint32 aLinAdr = RMmu::LinearAddress(aVal);
      console->Printf(_L("%08x -> Virtual  %08x\n"), aVal, aLinAdr);
    } else if (command == _L("e")) {
      if (lex.Val(aVal, EHex) == KErrNone) {
        exAdr = aVal;
      } else {
      	exAdr += 4;
      }
      aVal = RMmu::LinearAddress(exAdr);
      arlo.Examine(&aVal);
      console->Printf(_L("%08x: %08x\n"), exAdr, aVal);
    } else if (command == _L("x")) {
      isFinished = ETrue;
    } else if (command == _L("b")) {
      if (aPageIdx == 2) {
	console->Printf(_L("no file loaded\n"));
      } else {
	// boot
	TUint32 aPhysMapAdr = Linear2Physical(TUint32(aPageMap), arlo);
	// Boot: pass phys adr of map, and phys adr of code
	arlo.Boot(aPhysMapAdr, aPageMap[2]);
	// never come back here
      }
    } else if (command == _L("o")) {
      // copy option string to option buffer
      lex.SkipSpace();
      TPtrC anOpt = lex.Remainder();
      if (anOpt.Length() > 255) {
	console->Printf(_L("String is too long\n"));
      } else {
	TText8 *p = aOption;
	for (TInt i = 0; i < anOpt.Length(); i++)
	  *p++ = anOpt[i];
	*p++ = '\0';
      }
    } else if (command == _L("l")) {
      // load a file
      TPtrC bootName = lex.NextToken();
      loadFile(bootName, arlo);
    } else {
      console->Printf(_L("Unknown command\n"));
    }
  } while (!isFinished);
  // ok, that's all
  console->Printf(_L("Closing the channel\n"));
  arlo.Close();
  console->Printf(_L("Freeing the device\n"));
  User::LeaveIfError(User::FreeLogicalDevice(_L("ArLo")));
  // refers to device driver by loaded name - not filename
}

LOCAL_C void callExampleL() // initialize and call under cleanup stack
{
  console=Console::NewL(_L("ARLO"),
			TSize(KDefaultConsWidth,KDefaultConsHeight));
  CleanupStack::PushL(console);
  commandLine=CLineEdit::NewL(console,10);
  TInt fsret = fsSession.Connect(); // start a file session
  if (fsret != KErrNone) {
    console->Printf(_L("file session connect failed: return code=%d"),fsret);
    User::Leave(fsret);
  }
  TRAPD(error,doExampleL()); // perform example function
  if (error) {
    User::FreeLogicalDevice(_L("ArLo"));
    console->Printf(_L("failed: leave code=%d\n"), error);
    console->Printf(_L("Press a key\n"));
    console->Getch();
  }
  fsSession.Close(); //close the file session
  CleanupStack::PopAndDestroy(); // close console
}


GLDEF_C TInt E32Main() // main function called by E32
{
  __UHEAP_MARK;
  CTrapCleanup* cleanup=CTrapCleanup::New();       
  TRAPD(error,callExampleL());                     
  __ASSERT_ALWAYS(!error,User::Panic(_L("LINUX"),error));
  delete cleanup; // destroy clean-up stack
  __UHEAP_MARKEND;
  return 0; // and return
}

